@@ -1,4 +1,6 @@
-use crate::branch::{py_tag_selector, Branch, BranchOpenError, RegularBranch};
+use crate::branch::{py_tag_selector, Branch, RegularBranch};
+use crate::error::Error;
+use crate::repository::Repository;
 use crate::transport::Transport;
 use crate::tree::WorkingTree;
 
@@ -7,25 +9,7 @@ use crate::location::AsLocation;
 use pyo3::prelude::*;
 use pyo3::types::PyDict;
 
-pub struct Prober(PyObject);
-
-impl ToPyObject for Prober {
-    fn to_object(&self, py: Python) -> PyObject {
-        self.0.to_object(py)
-    }
-}
-
-impl FromPyObject<'_> for Prober {
-    fn extract(obj: &PyAny) -> PyResult<Self> {
-        Ok(Prober(obj.to_object(obj.py())))
-    }
-}
-
-impl Prober {
-    pub fn new(obj: PyObject) -> Self {
-        Prober(obj)
-    }
-}
+pub trait Prober: ToPyObject {}
 
 pub struct ControlDir(PyObject);
 
@@ -36,7 +20,7 @@ impl ToPyObject for ControlDir {
 }
 
 impl FromPyObject<'_> for ControlDir {
-    fn extract(obj: &PyAny) -> PyResult<Self> {
+    fn extract_bound(obj: &Bound<PyAny>) -> PyResult<Self> {
         Ok(ControlDir(obj.to_object(obj.py())))
     }
 }
@@ -54,8 +38,36 @@ impl ControlDir {
         .unwrap()
     }
 
+    pub fn user_transport(&self) -> Transport {
+        Python::with_gil(|py| {
+            let result = self.to_object(py).getattr(py, "user_transport").unwrap();
+            crate::transport::Transport::new(result)
+        })
+    }
+
+    pub fn control_transport(&self) -> Transport {
+        Python::with_gil(|py| {
+            let result = self.to_object(py).getattr(py, "control_transport").unwrap();
+            crate::transport::Transport::new(result)
+        })
+    }
+
+    pub fn open_repository(&self) -> Result<Repository, Error> {
+        Python::with_gil(|py| {
+            let result = self.to_object(py).call_method0(py, "open_repository")?;
+            Ok(Repository::new(result))
+        })
+    }
+
+    pub fn find_repository(&self) -> Result<Repository, Error> {
+        Python::with_gil(|py| {
+            let result = self.to_object(py).call_method0(py, "find_repository")?;
+            Ok(Repository::new(result))
+        })
+    }
+
     #[deprecated]
-    pub fn create_branch_convenience(base: &url::Url) -> Result<Box<dyn Branch>, CreateError> {
+    pub fn create_branch_convenience(base: &url::Url) -> Result<Box<dyn Branch>, Error> {
         create_branch_convenience(base)
     }
 
@@ -63,7 +75,7 @@ impl ControlDir {
     pub fn create_standalone_workingtree(
         base: &std::path::Path,
         format: impl AsFormat,
-    ) -> Result<WorkingTree, CreateError> {
+    ) -> Result<WorkingTree, Error> {
         create_standalone_workingtree(base, format)
     }
 
@@ -75,58 +87,65 @@ impl ControlDir {
         .unwrap()
     }
 
-    #[deprecated]
-    pub fn open_tree_or_branch(
-        location: &url::Url,
-        name: Option<&str>,
-    ) -> Result<(Option<WorkingTree>, Box<dyn Branch>), BranchOpenError> {
-        open_tree_or_branch(location, name, None)
-    }
-
-    #[deprecated]
-    pub fn open(url: &url::Url) -> Result<ControlDir, OpenError> {
-        open(url, None)
-    }
-
-    #[deprecated]
-    pub fn open_containing_from_transport(
-        transport: &Transport,
-        probers: Option<&[Prober]>,
-    ) -> Result<(ControlDir, String), OpenError> {
-        open_containing_from_transport(transport, probers)
-    }
-
-    #[deprecated]
-    pub fn open_from_transport(
-        transport: &Transport,
-        probers: Option<&[Prober]>,
-    ) -> Result<ControlDir, OpenError> {
-        open_from_transport(transport, probers)
-    }
-
-    pub fn create_branch(&self, name: Option<&str>) -> Result<Box<dyn Branch>, CreateError> {
+    pub fn create_branch(&self, name: Option<&str>) -> Result<Box<dyn Branch>, Error> {
         Python::with_gil(|py| {
             let branch = self
                 .to_object(py)
-                .call_method(py, "create_branch", (name,), None)?
+                .call_method_bound(py, "create_branch", (name,), None)?
                 .extract(py)?;
             Ok(Box::new(RegularBranch::new(branch)) as Box<dyn Branch>)
         })
     }
 
-    pub fn open_branch(
-        &self,
-        branch_name: Option<&str>,
-    ) -> Result<Box<dyn Branch>, BranchOpenError> {
+    pub fn create_repository(&self, shared: Option<bool>) -> Result<Repository, Error> {
+        Python::with_gil(|py| {
+            let kwargs = PyDict::new_bound(py);
+            if let Some(shared) = shared {
+                kwargs.set_item("shared", shared)?;
+            }
+            let repository = self
+                .to_object(py)
+                .call_method_bound(py, "create_repository", (), Some(&kwargs))?
+                .extract(py)?;
+            Ok(Repository::new(repository))
+        })
+    }
+
+    pub fn open_branch(&self, branch_name: Option<&str>) -> Result<Box<dyn Branch>, Error> {
         Python::with_gil(|py| {
             let branch = self
                 .to_object(py)
-                .call_method(py, "open_branch", (branch_name,), None)?
+                .call_method_bound(py, "open_branch", (branch_name,), None)?
                 .extract(py)?;
             Ok(Box::new(RegularBranch::new(branch)) as Box<dyn Branch>)
         })
     }
 
+    pub fn create_workingtree(&self) -> crate::Result<WorkingTree> {
+        Python::with_gil(|py| {
+            let wt = self
+                .to_object(py)
+                .call_method0(py, "create_workingtree")?
+                .extract(py)?;
+            Ok(WorkingTree(wt))
+        })
+    }
+
+    pub fn set_branch_reference(
+        &self,
+        branch: &dyn Branch,
+        name: Option<&str>,
+    ) -> crate::Result<()> {
+        Python::with_gil(|py| {
+            self.to_object(py).call_method1(
+                py,
+                "set_branch_reference",
+                (&branch.to_object(py), name),
+            )?;
+            Ok(())
+        })
+    }
+
     pub fn push_branch(
         &self,
         source_branch: &dyn Branch,
@@ -134,9 +153,9 @@ impl ControlDir {
         stop_revision: Option<&crate::RevisionId>,
         overwrite: Option<bool>,
         tag_selector: Option<Box<dyn Fn(String) -> bool>>,
-    ) -> PyResult<Box<dyn Branch>> {
+    ) -> crate::Result<Box<dyn Branch>> {
         Python::with_gil(|py| {
-            let kwargs = PyDict::new(py);
+            let kwargs = PyDict::new_bound(py);
             if let Some(to_branch_name) = to_branch_name {
                 kwargs.set_item("name", to_branch_name)?;
             }
@@ -149,11 +168,11 @@ impl ControlDir {
             if let Some(stop_revision) = stop_revision {
                 kwargs.set_item("stop_revision", stop_revision.to_object(py))?;
             }
-            let result = self.to_object(py).call_method(
+            let result = self.to_object(py).call_method_bound(
                 py,
                 "push_branch",
                 (&source_branch.to_object(py),),
-                Some(kwargs),
+                Some(&kwargs),
             )?;
             Ok(
                 Box::new(RegularBranch::new(result.getattr(py, "target_branch")?))
@@ -168,9 +187,10 @@ impl ControlDir {
         source_branch: Option<&dyn Branch>,
         create_tree_if_local: Option<bool>,
         stacked: Option<bool>,
-    ) -> ControlDir {
+        revision_id: Option<&crate::RevisionId>,
+    ) -> Result<ControlDir, Error> {
         Python::with_gil(|py| {
-            let kwargs = PyDict::new(py);
+            let kwargs = PyDict::new_bound(py);
             if let Some(create_tree_if_local) = create_tree_if_local {
                 kwargs
                     .set_item("create_tree_if_local", create_tree_if_local)
@@ -181,26 +201,40 @@ impl ControlDir {
             }
             if let Some(source_branch) = source_branch {
                 kwargs
-                    .set_item("source_branch", &source_branch.to_object(py))
+                    .set_item("source_branch", source_branch.to_object(py))
+                    .unwrap();
+            }
+            if let Some(revision_id) = revision_id {
+                kwargs
+                    .set_item("revision_id", revision_id.to_object(py))
                     .unwrap();
             }
 
-            let cd = self
-                .0
-                .call_method(py, "sprout", (target.to_string(),), Some(kwargs))
-                .unwrap();
-            ControlDir(cd)
+            let cd =
+                self.0
+                    .call_method_bound(py, "sprout", (target.to_string(),), Some(&kwargs))?;
+            Ok(ControlDir(cd))
         })
     }
 
-    pub fn open_workingtree(&self) -> PyResult<WorkingTree> {
+    pub fn has_workingtree(&self) -> bool {
+        Python::with_gil(|py| {
+            let result = self
+                .to_object(py)
+                .call_method0(py, "has_workingtree")
+                .unwrap();
+            result.extract(py).unwrap()
+        })
+    }
+
+    pub fn open_workingtree(&self) -> crate::Result<WorkingTree> {
         Python::with_gil(|py| {
             let wt = self.0.call_method0(py, "open_workingtree")?.extract(py)?;
             Ok(WorkingTree(wt))
         })
     }
 
-    pub fn branch_names(&self) -> PyResult<Vec<String>> {
+    pub fn branch_names(&self) -> crate::Result<Vec<String>> {
         Python::with_gil(|py| {
             let names = self
                 .0
@@ -240,7 +274,7 @@ impl From<PyObject> for ControlDirFormat {
 impl Default for ControlDirFormat {
     fn default() -> Self {
         Python::with_gil(|py| {
-            let breezy = PyModule::import(py, "breezy.controldir").unwrap();
+            let breezy = PyModule::import_bound(py, "breezy.controldir").unwrap();
             let cd_format = breezy.getattr("ControlDirFormat").unwrap();
             let obj = cd_format.call_method0("get_default_format").unwrap();
             assert!(!obj.is_none());
@@ -285,99 +319,45 @@ impl ControlDirFormat {
                 .unwrap()
         })
     }
-}
 
-#[derive(Debug)]
-pub enum OpenError {
-    Python(PyErr),
-    NotFound(String),
-    UnknownFormat,
-}
+    pub fn initialize_on_transport(&self, transport: &Transport) -> Result<ControlDir, Error> {
+        Python::with_gil(|py| {
+            let cd =
+                self.0
+                    .call_method1(py, "initialize_on_transport", (&transport.to_object(py),))?;
+            Ok(ControlDir(cd))
+        })
+    }
 
-impl From<PyErr> for OpenError {
-    fn from(err: PyErr) -> Self {
-        pyo3::import_exception!(breezy.errors, NotBranchError);
-        pyo3::import_exception!(breezy.errors, UnknownFormatError);
-
-        pyo3::Python::with_gil(|py| {
-            if err.is_instance_of::<NotBranchError>(py) {
-                OpenError::NotFound(err.value(py).getattr("path").unwrap().extract().unwrap())
-            } else if err.is_instance_of::<UnknownFormatError>(py) {
-                OpenError::UnknownFormat
-            } else {
-                OpenError::Python(err)
-            }
+    pub fn initialize(&self, location: impl AsLocation) -> Result<ControlDir, Error> {
+        Python::with_gil(|py| {
+            let cd = self
+                .0
+                .call_method1(py, "initialize", (location.as_location(),))?;
+            Ok(ControlDir(cd))
         })
     }
 }
 
-impl std::fmt::Display for OpenError {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            OpenError::Python(err) => err.fmt(f),
-            OpenError::NotFound(name) => write!(f, "Not found: {}", name),
-            OpenError::UnknownFormat => write!(f, "Unknown format"),
-        }
-    }
-}
-
-impl std::error::Error for OpenError {}
-
-#[derive(Debug)]
-pub enum CreateError {
-    Python(PyErr),
-    AlreadyExists,
-    UnknownFormat,
-}
-
-impl From<PyErr> for CreateError {
-    fn from(err: PyErr) -> Self {
-        pyo3::import_exception!(breezy.errors, AlreadyControlDirError);
-        pyo3::import_exception!(breezy.errors, UnknownFormatError);
-
-        pyo3::Python::with_gil(|py| {
-            if err.is_instance_of::<AlreadyControlDirError>(py) {
-                CreateError::AlreadyExists
-            } else if err.is_instance_of::<UnknownFormatError>(py) {
-                CreateError::UnknownFormat
-            } else {
-                CreateError::Python(err)
-            }
-        })
-    }
-}
-
-impl std::fmt::Display for CreateError {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            CreateError::Python(err) => err.fmt(f),
-            CreateError::AlreadyExists => write!(f, "Already exists"),
-            CreateError::UnknownFormat => write!(f, "Unknown format"),
-        }
-    }
-}
-
-impl std::error::Error for CreateError {}
-
 pub fn open_tree_or_branch(
     location: impl AsLocation,
     name: Option<&str>,
     possible_transports: Option<&mut Vec<Transport>>,
-) -> Result<(Option<WorkingTree>, Box<dyn Branch>), BranchOpenError> {
+) -> Result<(Option<WorkingTree>, Box<dyn Branch>), Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
 
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         if let Some(possible_transports) = possible_transports {
             kwargs.set_item("possible_transports", possible_transports.to_object(py))?;
         }
 
-        let ret = cd.to_object(py).call_method(
+        let ret = cd.to_object(py).call_method_bound(
             py,
             "open_tree_or_branch",
             (location.as_location(), name),
-            Some(kwargs),
+            Some(&kwargs),
         )?;
 
         let (tree, branch) = ret.extract::<(Option<PyObject>, PyObject)>(py)?;
@@ -397,7 +377,11 @@ fn test_open_tree_or_branch() {
         None,
     )
     .unwrap();
-    assert_eq!(wt.unwrap().basedir(), tmp_dir.path());
+
+    assert_eq!(
+        wt.unwrap().basedir().canonicalize().unwrap(),
+        tmp_dir.path().canonicalize().unwrap()
+    );
     assert_eq!(
         branch.get_user_url(),
         url::Url::from_directory_path(tmp_dir.path()).unwrap()
@@ -407,15 +391,15 @@ fn test_open_tree_or_branch() {
 pub fn open(
     url: impl AsLocation,
     possible_transports: Option<&mut Vec<Transport>>,
-) -> Result<ControlDir, OpenError> {
+) -> Result<ControlDir, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         if let Some(possible_transports) = possible_transports {
             kwargs.set_item("possible_transports", possible_transports.to_object(py))?;
         }
-        let controldir = cd.call_method("open", (url.as_location(),), Some(kwargs))?;
+        let controldir = cd.call_method("open", (url.as_location(),), Some(&kwargs))?;
         Ok(ControlDir(controldir.to_object(py)))
     })
 }
@@ -430,7 +414,7 @@ fn test_open() {
     )
     .unwrap_err();
 
-    assert!(matches!(e, OpenError::NotFound(_)),);
+    assert!(matches!(e, Error::NotBranchError(..)),);
 
     let cd = create(
         &url::Url::from_directory_path(tmp_dir.path()).unwrap(),
@@ -454,18 +438,18 @@ pub fn create(
     url: impl AsLocation,
     format: impl AsFormat,
     possible_transports: Option<&mut Vec<Transport>>,
-) -> Result<ControlDir, CreateError> {
+) -> Result<ControlDir, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         if let Some(format) = format.as_format() {
             kwargs.set_item("format", format.to_object(py))?;
         }
         if let Some(possible_transports) = possible_transports {
             kwargs.set_item("possible_transports", possible_transports.to_object(py))?;
         }
-        let controldir = cd.call_method("create", (url.as_location(),), Some(kwargs))?;
+        let controldir = cd.call_method("create", (url.as_location(),), Some(&kwargs))?;
         Ok(ControlDir(controldir.to_object(py)))
     })
 }
@@ -494,10 +478,10 @@ fn test_create() {
 pub fn create_on_transport(
     transport: &Transport,
     format: impl AsFormat,
-) -> Result<ControlDir, CreateError> {
+) -> Result<ControlDir, Error> {
     Python::with_gil(|py| {
         let format = format.as_format().unwrap().0;
-        Ok(ControlDir(format.call_method(
+        Ok(ControlDir(format.call_method_bound(
             py,
             "initialize_on_transport",
             (&transport.to_object(py),),
@@ -519,21 +503,24 @@ fn test_create_on_transport() {
 
 pub fn open_containing_from_transport(
     transport: &Transport,
-    probers: Option<&[Prober]>,
-) -> Result<(ControlDir, String), OpenError> {
+    probers: Option<&[&dyn Prober]>,
+) -> Result<(ControlDir, String), Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         if let Some(probers) = probers {
-            kwargs.set_item("probers", probers.iter().map(|p| &p.0).collect::<Vec<_>>())?;
+            kwargs.set_item(
+                "probers",
+                probers.iter().map(|p| p.to_object(py)).collect::<Vec<_>>(),
+            )?;
         }
 
         let (controldir, subpath): (PyObject, String) = cd
             .call_method(
                 "open_containing_from_transport",
                 (&transport.to_object(py),),
-                Some(kwargs),
+                Some(&kwargs),
             )?
             .extract()?;
         Ok((ControlDir(controldir.to_object(py)), subpath))
@@ -549,24 +536,27 @@ fn test_open_containing_from_transport() {
     )
     .unwrap();
     let e = open_containing_from_transport(&transport, None).unwrap_err();
-    assert!(matches!(e, OpenError::NotFound(_)),);
+    assert!(matches!(e, Error::NotBranchError(..)),);
 }
 
 pub fn open_from_transport(
     transport: &Transport,
-    probers: Option<&[Prober]>,
-) -> Result<ControlDir, OpenError> {
+    probers: Option<&[&dyn Prober]>,
+) -> Result<ControlDir, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         if let Some(probers) = probers {
-            kwargs.set_item("probers", probers.iter().map(|p| &p.0).collect::<Vec<_>>())?;
+            kwargs.set_item(
+                "probers",
+                probers.iter().map(|p| p.to_object(py)).collect::<Vec<_>>(),
+            )?;
         }
         let controldir = cd.call_method(
             "open_from_transport",
             (&transport.to_object(py),),
-            Some(kwargs),
+            Some(&kwargs),
         )?;
         Ok(ControlDir(controldir.to_object(py)))
     })
@@ -581,7 +571,7 @@ fn test_open_from_transport() {
     )
     .unwrap();
     let e = open_from_transport(&transport, None).unwrap_err();
-    assert!(matches!(e, OpenError::NotFound(_)),);
+    assert!(matches!(e, Error::NotBranchError(..)),);
 }
 
 pub trait AsFormat {
@@ -591,7 +581,7 @@ pub trait AsFormat {
 impl AsFormat for &str {
     fn as_format(&self) -> Option<ControlDirFormat> {
         Python::with_gil(|py| {
-            let m = py.import("breezy.controldir").ok()?;
+            let m = py.import_bound("breezy.controldir").ok()?;
             let cd = m.getattr("format_registry").ok()?;
             let format = cd
                 .call_method1("make_controldir", (self.to_string(),))
@@ -607,9 +597,9 @@ impl AsFormat for &ControlDirFormat {
     }
 }
 
-pub fn create_branch_convenience(base: &url::Url) -> Result<Box<dyn Branch>, CreateError> {
+pub fn create_branch_convenience(base: &url::Url) -> Result<Box<dyn Branch>, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
         let branch = cd.call_method("create_branch_convenience", (base.to_string(),), None)?;
         Ok(Box::new(RegularBranch::new(branch.to_object(py))) as Box<dyn Branch>)
@@ -624,10 +614,10 @@ pub fn create_branch_convenience(base: &url::Url) -> Result<Box<dyn Branch>, Cre
 pub fn create_standalone_workingtree(
     base: &std::path::Path,
     format: impl AsFormat,
-) -> Result<WorkingTree, CreateError> {
+) -> Result<WorkingTree, Error> {
     let base = base.to_str().unwrap();
     Python::with_gil(|py| {
-        let m = py.import("breezy.controldir")?;
+        let m = py.import_bound("breezy.controldir")?;
         let cd = m.getattr("ControlDir")?;
         let format = format.as_format();
         let wt = cd.call_method(
@@ -644,7 +634,10 @@ fn test_create_standalone_workingtree() {
     let tmp_dir = tempfile::tempdir().unwrap();
     let wt = create_standalone_workingtree(tmp_dir.path(), "2a").unwrap();
 
-    assert_eq!(wt.basedir(), tmp_dir.path());
+    assert_eq!(
+        wt.basedir().canonicalize().unwrap(),
+        tmp_dir.path().canonicalize().unwrap()
+    );
 }
 
 #[test]
@@ -658,3 +651,136 @@ fn test_create_branch_convenience() {
         url::Url::from_directory_path(tmp_dir.path()).unwrap()
     );
 }
+
+pub fn all_probers() -> Vec<Box<dyn Prober>> {
+    struct PyProber(PyObject);
+    impl ToPyObject for PyProber {
+        fn to_object(&self, py: Python) -> PyObject {
+            self.0.clone_ref(py)
+        }
+    }
+    impl Prober for PyProber {}
+    Python::with_gil(|py| -> PyResult<Vec<Box<dyn Prober>>> {
+        let m = py.import_bound("breezy.controldir")?;
+        let cdf = m.getattr("ControlDirFormat")?;
+        let probers = cdf.call_method0("all_probers")?.extract::<Vec<PyObject>>()?;
+        Ok(probers
+            .into_iter()
+            .map(|p| Box::new(PyProber(p)) as Box<dyn Prober>)
+            .collect::<Vec<_>>())
+    })
+    .unwrap()
+}
+
+#[test]
+fn test_create_repository() {
+    let tmp_dir = tempfile::tempdir().unwrap();
+    let controldir = create(
+        &url::Url::from_directory_path(tmp_dir.path()).unwrap(),
+        &ControlDirFormat::default(),
+        None,
+    )
+    .unwrap();
+    let _repo = controldir.create_repository(None).unwrap();
+}
+
+#[test]
+fn test_create_branch() {
+    let tmp_dir = tempfile::tempdir().unwrap();
+    let controldir = create(
+        &url::Url::from_directory_path(tmp_dir.path()).unwrap(),
+        &ControlDirFormat::default(),
+        None,
+    )
+    .unwrap();
+    assert!(matches!(
+        controldir.create_branch(None),
+        Err(Error::NoRepositoryPresent)
+    ));
+    let _repo = controldir.create_repository(None).unwrap();
+    let _branch = controldir.create_branch(Some("foo")).unwrap();
+}
+
+#[test]
+fn test_create_workingtree() {
+    let tmp_dir = tempfile::tempdir().unwrap();
+    let controldir = create(
+        &url::Url::from_directory_path(tmp_dir.path()).unwrap(),
+        &ControlDirFormat::default(),
+        None,
+    )
+    .unwrap();
+    controldir.create_repository(None).unwrap();
+    controldir.create_branch(None).unwrap();
+    let _wt = controldir.create_workingtree().unwrap();
+}
+
+#[test]
+fn test_branch_names() {
+    let tmp_dir = tempfile::tempdir().unwrap();
+    let controldir = create(
+        &url::Url::from_directory_path(tmp_dir.path()).unwrap(),
+        &ControlDirFormat::default(),
+        None,
+    )
+    .unwrap();
+    controldir.create_repository(None).unwrap();
+    controldir.create_branch(None).unwrap();
+    assert_eq!(controldir.branch_names().unwrap(), vec!["".to_string()]);
+}
+
+pub struct ControlDirFormatRegistry(PyObject);
+
+impl ControlDirFormatRegistry {
+    pub fn new() -> Self {
+        Python::with_gil(|py| {
+            let m = py.import_bound("breezy.controldir").unwrap();
+            let obj = m.getattr("format_registry").unwrap();
+            ControlDirFormatRegistry(obj.into())
+        })
+    }
+
+    pub fn make_controldir(&self, format: &str) -> Option<ControlDirFormat> {
+        Python::with_gil(
+            |py| match self.0.call_method1(py, "make_controldir", (format,)) {
+                Ok(format) => Some(ControlDirFormat(format.to_object(py))),
+                Err(e) if e.is_instance_of::<pyo3::exceptions::PyKeyError>(py) => None,
+                Err(e) => panic!("{}", e),
+            },
+        )
+    }
+}
+
+impl Default for ControlDirFormatRegistry {
+    fn default() -> Self {
+        ControlDirFormatRegistry::new()
+    }
+}
+
+lazy_static::lazy_static! {
+    pub static ref FORMAT_REGISTRY: ControlDirFormatRegistry = ControlDirFormatRegistry::new();
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_control_dir_format_registry() {
+        let registry = ControlDirFormatRegistry::new();
+        let format = registry.make_controldir("2a").unwrap();
+        let _ = format.get_format_string();
+    }
+
+    #[test]
+    fn test_format_registry() {
+        let format = FORMAT_REGISTRY.make_controldir("2a").unwrap();
+        let _ = format.get_format_string();
+    }
+
+    #[test]
+    fn test_all_probers() {
+        let probers = all_probers();
+        assert!(!probers.is_empty());
+    }
+}
