@@ -1,73 +1,11 @@
 use crate::branch::{py_tag_selector, Branch, RegularBranch};
+use crate::error::Error;
 use crate::revisionid::RevisionId;
 use pyo3::conversion::ToPyObject;
 use pyo3::exceptions::PyValueError;
-use pyo3::import_exception;
 use pyo3::prelude::*;
 use pyo3::types::PyDict;
-
-import_exception!(breezy.forge, ForgeLoginRequired);
-import_exception!(breezy.forge, UnsupportedForge);
-import_exception!(breezy.errors, AlreadyControlDirError);
-
-#[derive(Clone, Debug)]
-pub enum Error {
-    LoginRequired,
-    UnsupportedForge(url::Url),
-    ProjectExists(String),
-}
-
-impl std::fmt::Display for Error {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            Error::LoginRequired => write!(f, "Login required"),
-            Error::UnsupportedForge(url) => write!(f, "Unsupported forge: {}", url),
-            Error::ProjectExists(name) => write!(f, "Project already exists: {}", name),
-        }
-    }
-}
-
-impl std::error::Error for Error {}
-
-impl From<PyErr> for Error {
-    fn from(err: PyErr) -> Self {
-        Python::with_gil(|py| {
-            if err.is_instance_of::<ForgeLoginRequired>(py) {
-                Error::LoginRequired
-            } else if err.is_instance_of::<UnsupportedForge>(py) {
-                Error::UnsupportedForge(
-                    err.value(py)
-                        .getattr("branch")
-                        .unwrap()
-                        .extract::<String>()
-                        .unwrap()
-                        .parse()
-                        .unwrap(),
-                )
-            } else if err.is_instance_of::<AlreadyControlDirError>(py) {
-                Error::ProjectExists(
-                    err.value(py)
-                        .getattr("path")
-                        .unwrap()
-                        .extract::<String>()
-                        .unwrap(),
-                )
-            } else {
-                panic!("Unexpected error: {}", err);
-            }
-        })
-    }
-}
-
-impl From<Error> for PyErr {
-    fn from(err: Error) -> PyErr {
-        match err {
-            Error::LoginRequired => ForgeLoginRequired::new_err("Login required"),
-            Error::UnsupportedForge(url) => UnsupportedForge::new_err(url.to_string()),
-            Error::ProjectExists(name) => AlreadyControlDirError::new_err(name),
-        }
-    }
-}
+use std::hash::Hash;
 
 #[derive(Clone)]
 pub struct Forge(PyObject);
@@ -78,7 +16,25 @@ impl From<PyObject> for Forge {
     }
 }
 
-#[derive(Clone, Copy)]
+impl std::fmt::Debug for MergeProposal {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        // Just print the URL for now
+        let mut s = f.debug_struct("MergeProposal");
+        if let Ok(url) = self.url() {
+            s.field("url", &url);
+        }
+        s.finish()
+    }
+}
+
+impl std::fmt::Display for MergeProposal {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        let url = self.url().unwrap();
+        write!(f, "{}", url)
+    }
+}
+
+#[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
 pub enum MergeProposalStatus {
     All,
     Open,
@@ -106,13 +62,13 @@ impl std::str::FromStr for MergeProposalStatus {
     }
 }
 
-impl ToString for MergeProposalStatus {
-    fn to_string(&self) -> String {
+impl std::fmt::Display for MergeProposalStatus {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         match self {
-            MergeProposalStatus::All => "all".to_string(),
-            MergeProposalStatus::Open => "open".to_string(),
-            MergeProposalStatus::Merged => "merged".to_string(),
-            MergeProposalStatus::Closed => "closed".to_string(),
+            MergeProposalStatus::All => write!(f, "all"),
+            MergeProposalStatus::Open => write!(f, "open"),
+            MergeProposalStatus::Merged => write!(f, "merged"),
+            MergeProposalStatus::Closed => write!(f, "closed"),
         }
     }
 }
@@ -124,7 +80,7 @@ impl ToPyObject for MergeProposalStatus {
 }
 
 impl FromPyObject<'_> for MergeProposalStatus {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<PyAny>) -> PyResult<Self> {
         let status = ob.extract::<String>()?;
         match status.as_str() {
             "all" => Ok(MergeProposalStatus::All),
@@ -149,63 +105,70 @@ impl From<PyObject> for MergeProposal {
 }
 
 impl MergeProposal {
-    pub fn reopen(&self) -> PyResult<()> {
+    pub fn from_url(url: &url::Url) -> Result<Self, Error> {
+        get_proposal_by_url(url)
+    }
+
+    pub fn reopen(&self) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0.call_method0(py, "reopen")?;
             Ok(())
         })
     }
 
-    pub fn close(&self) -> PyResult<()> {
+    pub fn close(&self) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0.call_method0(py, "close")?;
             Ok(())
         })
     }
 
-    pub fn url(&self) -> PyResult<url::Url> {
+    pub fn url(&self) -> Result<url::Url, crate::error::Error> {
         Python::with_gil(|py| {
             let url = self.0.getattr(py, "url")?;
             Ok(url.extract::<String>(py)?.parse().unwrap())
         })
     }
 
-    pub fn is_merged(&self) -> PyResult<bool> {
+    pub fn is_merged(&self) -> Result<bool, crate::error::Error> {
         Python::with_gil(|py| {
             let is_merged = self.0.call_method0(py, "is_merged")?;
-            is_merged.extract(py)
+            is_merged.extract(py).map_err(|e| e.into())
         })
     }
 
-    pub fn is_closed(&self) -> PyResult<bool> {
+    pub fn is_closed(&self) -> Result<bool, crate::error::Error> {
         Python::with_gil(|py| {
             let is_closed = self.0.call_method0(py, "is_closed")?;
-            is_closed.extract(py)
+            is_closed.extract(py).map_err(|e| e.into())
         })
     }
 
-    pub fn get_title(&self) -> PyResult<Option<String>> {
+    pub fn get_title(&self) -> Result<Option<String>, crate::error::Error> {
         Python::with_gil(|py| {
             let title = self.0.call_method0(py, "get_title")?;
-            title.extract(py)
+            title.extract(py).map_err(|e| e.into())
         })
     }
 
-    pub fn set_title(&self, title: Option<&str>) -> PyResult<()> {
+    pub fn set_title(&self, title: Option<&str>) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0.call_method1(py, "set_title", (title,))?;
             Ok(())
         })
     }
 
-    pub fn get_commit_message(&self) -> PyResult<Option<String>> {
+    pub fn get_commit_message(&self) -> Result<Option<String>, crate::error::Error> {
         Python::with_gil(|py| {
             let commit_message = self.0.call_method0(py, "get_commit_message")?;
-            commit_message.extract(py)
+            commit_message.extract(py).map_err(|e| e.into())
         })
     }
 
-    pub fn set_commit_message(&self, commit_message: Option<&str>) -> PyResult<()> {
+    pub fn set_commit_message(
+        &self,
+        commit_message: Option<&str>,
+    ) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0
                 .call_method1(py, "set_commit_message", (commit_message,))?;
@@ -213,46 +176,46 @@ impl MergeProposal {
         })
     }
 
-    pub fn get_target_branch_url(&self) -> PyResult<Option<url::Url>> {
+    pub fn get_target_branch_url(&self) -> Result<Option<url::Url>, crate::error::Error> {
         Python::with_gil(|py| {
             let target_branch_url = self.0.call_method0(py, "get_target_branch_url")?;
             target_branch_url
                 .extract::<String>(py)?
-                .parse()
+                .parse::<url::Url>()
                 .map(Some)
-                .map_err(|e| PyValueError::new_err(format!("{}", e)))
+                .map_err(|e| e.into())
         })
     }
 
-    pub fn get_source_branch_url(&self) -> PyResult<Option<url::Url>> {
+    pub fn get_source_branch_url(&self) -> Result<Option<url::Url>, crate::error::Error> {
         Python::with_gil(|py| {
             let source_branch_url = self.0.call_method0(py, "get_source_branch_url")?;
             source_branch_url
                 .extract::<String>(py)?
-                .parse()
+                .parse::<url::Url>()
                 .map(Some)
-                .map_err(|e| PyValueError::new_err(format!("{}", e)))
+                .map_err(|e| e.into())
         })
     }
 
-    pub fn get_description(&self) -> PyResult<Option<String>> {
+    pub fn get_description(&self) -> Result<Option<String>, crate::error::Error> {
         Python::with_gil(|py| {
             let description = self.0.call_method0(py, "get_description")?;
-            description.extract(py)
+            description.extract(py).map_err(|e| e.into())
         })
     }
 
-    pub fn set_description(&self, description: Option<&str>) -> PyResult<()> {
+    pub fn set_description(&self, description: Option<&str>) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0.call_method1(py, "set_description", (description,))?;
             Ok(())
         })
     }
 
-    pub fn can_be_merged(&self) -> PyResult<bool> {
+    pub fn can_be_merged(&self) -> Result<bool, crate::error::Error> {
         Python::with_gil(|py| {
             let can_be_merged = self.0.call_method0(py, "can_be_merged")?;
-            can_be_merged.extract(py)
+            can_be_merged.extract(py).map_err(|e| e.into())
         })
     }
 
@@ -272,6 +235,16 @@ impl MergeProposal {
             Ok(())
         })
     }
+
+    pub fn get_web_url(&self) -> Result<url::Url, crate::error::Error> {
+        Python::with_gil(|py| {
+            let web_url = self.0.call_method0(py, "get_web_url")?;
+            web_url
+                .extract::<String>(py)?
+                .parse::<url::Url>()
+                .map_err(|e| e.into())
+        })
+    }
 }
 
 #[pyclass]
@@ -281,7 +254,7 @@ impl ProposalBuilder {
     pub fn description(self, description: &str) -> Self {
         Python::with_gil(|py| {
             self.1
-                .as_ref(py)
+                .bind(py)
                 .set_item("description", description)
                 .unwrap();
         });
@@ -290,14 +263,14 @@ impl ProposalBuilder {
 
     pub fn labels(self, labels: &[&str]) -> Self {
         Python::with_gil(|py| {
-            self.1.as_ref(py).set_item("labels", labels).unwrap();
+            self.1.bind(py).set_item("labels", labels).unwrap();
         });
         self
     }
 
     pub fn reviewers(self, reviewers: &[&str]) -> Self {
         Python::with_gil(|py| {
-            self.1.as_ref(py).set_item("reviewers", reviewers).unwrap();
+            self.1.bind(py).set_item("reviewers", reviewers).unwrap();
         });
         self
     }
@@ -305,7 +278,7 @@ impl ProposalBuilder {
     pub fn allow_collaboration(self, allow_collaboration: bool) -> Self {
         Python::with_gil(|py| {
             self.1
-                .as_ref(py)
+                .bind(py)
                 .set_item("allow_collaboration", allow_collaboration)
                 .unwrap();
         });
@@ -314,7 +287,7 @@ impl ProposalBuilder {
 
     pub fn title(self, title: &str) -> Self {
         Python::with_gil(|py| {
-            self.1.as_ref(py).set_item("title", title).unwrap();
+            self.1.bind(py).set_item("title", title).unwrap();
         });
         self
     }
@@ -322,14 +295,14 @@ impl ProposalBuilder {
     pub fn commit_message(self, commit_message: &str) -> Self {
         Python::with_gil(|py| {
             self.1
-                .as_ref(py)
+                .bind(py)
                 .set_item("commit_message", commit_message)
                 .unwrap();
         });
         self
     }
 
-    pub fn build(self) -> PyResult<MergeProposal> {
+    pub fn build(self) -> Result<MergeProposal, crate::error::Error> {
         Python::with_gil(|py| {
             let kwargs = self.1;
             let proposal = self.0.call_method1(py, "create_proposal", (kwargs,))?;
@@ -339,7 +312,10 @@ impl ProposalBuilder {
 }
 
 impl Forge {
-    pub fn get_proposal_by_url(&self, url: &url::Url) -> PyResult<MergeProposal> {
+    pub fn get_proposal_by_url(
+        &self,
+        url: &url::Url,
+    ) -> Result<MergeProposal, crate::error::Error> {
         Python::with_gil(|py| {
             let proposal =
                 self.to_object(py)
@@ -348,6 +324,18 @@ impl Forge {
         })
     }
 
+    pub fn get_web_url(&self, branch: &dyn Branch) -> Result<url::Url, crate::error::Error> {
+        Python::with_gil(|py| {
+            let url = self
+                .to_object(py)
+                .call_method1(py, "get_web_url", (&branch.to_object(py),))
+                ?
+                .extract::<String>(py)
+                .unwrap();
+            Ok(url.parse::<url::Url>().unwrap())
+        })
+    }
+
     pub fn base_url(&self) -> url::Url {
         Python::with_gil(|py| {
             let base_url = self.to_object(py).getattr(py, "base_url").unwrap();
@@ -358,7 +346,7 @@ impl Forge {
     pub fn forge_kind(&self) -> String {
         Python::with_gil(|py| {
             self.to_object(py)
-                .as_ref(py)
+                .bind(py)
                 .get_type()
                 .name()
                 .unwrap()
@@ -369,7 +357,7 @@ impl Forge {
     pub fn forge_name(&self) -> String {
         Python::with_gil(|py| {
             self.to_object(py)
-                .as_ref(py)
+                .bind(py)
                 .get_type()
                 .name()
                 .unwrap()
@@ -407,11 +395,21 @@ impl Forge {
         })
     }
 
+    pub fn supports_merge_proposal_labels(&self) -> bool {
+        Python::with_gil(|py| {
+            let supports_merge_proposal_labels = self
+                .to_object(py)
+                .getattr(py, "supports_merge_proposal_labels")
+                .unwrap();
+            supports_merge_proposal_labels.extract(py).unwrap()
+        })
+    }
+
     pub fn get_proposer(
         &self,
         from_branch: &dyn Branch,
         to_branch: &dyn Branch,
-    ) -> PyResult<ProposalBuilder> {
+    ) -> Result<ProposalBuilder, crate::error::Error> {
         Python::with_gil(|py| {
             Ok(ProposalBuilder(
                 self.0.call_method1(
@@ -419,7 +417,7 @@ impl Forge {
                     "get_proposer",
                     (from_branch.to_object(py), to_branch.to_object(py)),
                 )?,
-                PyDict::new(py).into(),
+                PyDict::new_bound(py).into(),
             ))
         })
     }
@@ -433,13 +431,13 @@ impl Forge {
             Python::with_gil(|py| -> Result<Vec<MergeProposal>, Error> {
                 Ok(self
                     .to_object(py)
-                    .call_method(
+                    .call_method_bound(
                         py,
                         "iter_my_proposals",
                         (status.to_object(py), author),
                         None,
                     )?
-                    .as_ref(py)
+                    .bind(py)
                     .iter()
                     .unwrap()
                     .map(|proposal| MergeProposal::from(proposal.unwrap().to_object(py)))
@@ -454,20 +452,20 @@ impl Forge {
         name: &str,
         owner: Option<&str>,
         preferred_schemes: Option<&[&str]>,
-    ) -> PyResult<Box<dyn Branch>> {
+    ) -> Result<Box<dyn Branch>, crate::error::Error> {
         Python::with_gil(|py| {
-            let kwargs = PyDict::new(py);
+            let kwargs = PyDict::new_bound(py);
             if let Some(owner) = owner {
                 kwargs.set_item("owner", owner)?;
             }
             if let Some(preferred_schemes) = preferred_schemes {
                 kwargs.set_item("preferred_schemes", preferred_schemes)?;
             }
-            let branch = self.to_object(py).call_method(
+            let branch = self.to_object(py).call_method_bound(
                 py,
                 "get_derived_branch",
                 (main_branch.to_object(py), name),
-                Some(kwargs),
+                Some(&kwargs),
             )?;
             Ok(Box::new(RegularBranch::new(branch)) as Box<dyn Branch>)
         })
@@ -478,17 +476,17 @@ impl Forge {
         source_branch: &dyn Branch,
         target_branch: &dyn Branch,
         status: MergeProposalStatus,
-    ) -> PyResult<impl Iterator<Item = MergeProposal>> {
+    ) -> Result<impl Iterator<Item = MergeProposal>, crate::error::Error> {
         Python::with_gil(move |py| {
-            let kwargs = PyDict::new(py);
+            let kwargs = PyDict::new_bound(py);
             kwargs.set_item("status", status.to_string())?;
             let proposals: Vec<PyObject> = self
                 .0
-                .call_method(
+                .call_method_bound(
                     py,
                     "iter_proposals",
                     (&source_branch.to_object(py), &target_branch.to_object(py)),
-                    Some(kwargs),
+                    Some(&kwargs),
                 )?
                 .extract(py)?;
             Ok(proposals.into_iter().map(MergeProposal::from))
@@ -504,11 +502,11 @@ impl Forge {
         owner: Option<&str>,
         stop_revision: Option<&RevisionId>,
         tag_selector: Option<Box<dyn Fn(String) -> bool>>,
-    ) -> PyResult<(Box<dyn Branch>, url::Url)> {
+    ) -> Result<(Box<dyn Branch>, url::Url), crate::error::Error> {
         Python::with_gil(|py| {
-            let kwargs = PyDict::new(py);
-            kwargs.set_item("local_branch", &local_branch.to_object(py))?;
-            kwargs.set_item("main_branch", &main_branch.to_object(py))?;
+            let kwargs = PyDict::new_bound(py);
+            kwargs.set_item("local_branch", local_branch.to_object(py))?;
+            kwargs.set_item("main_branch", main_branch.to_object(py))?;
             kwargs.set_item("name", name)?;
             if let Some(overwrite_existing) = overwrite_existing {
                 kwargs.set_item("overwrite_existing", overwrite_existing)?;
@@ -524,7 +522,7 @@ impl Forge {
             }
             let (b, u): (PyObject, String) = self
                 .to_object(py)
-                .call_method(py, "publish_derived", (), Some(kwargs))?
+                .call_method_bound(py, "publish_derived", (), Some(&kwargs))?
                 .extract(py)?;
             Ok((
                 Box::new(RegularBranch::new(b)) as Box<dyn Branch>,
@@ -548,12 +546,23 @@ impl Forge {
 
 impl std::fmt::Debug for Forge {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        f.debug_struct("Forge").finish()
+        let mut s = f.debug_struct("Forge");
+        if let Ok(base_url) = self.base_url().to_string().parse::<url::Url>() {
+            s.field("base_url", &base_url);
+        }
+        s.finish()
+    }
+}
+
+impl std::fmt::Display for Forge {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        let base_url = self.base_url();
+        write!(f, "{}", base_url)
     }
 }
 
 impl FromPyObject<'_> for Forge {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<PyAny>) -> PyResult<Self> {
         Ok(Forge(ob.to_object(ob.py())))
     }
 }
@@ -566,7 +575,7 @@ impl ToPyObject for Forge {
 
 pub fn get_forge(branch: &dyn Branch) -> Result<Forge, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.forge").unwrap();
+        let m = py.import_bound("breezy.forge").unwrap();
         let forge = m.call_method1("get_forge", (branch.to_object(py),))?;
         Ok(Forge(forge.to_object(py)))
     })
@@ -574,7 +583,7 @@ pub fn get_forge(branch: &dyn Branch) -> Result<Forge, Error> {
 
 pub fn determine_title(description: &str) -> String {
     Python::with_gil(|py| {
-        let m = py.import("breezy.forge").unwrap();
+        let m = py.import_bound("breezy.forge").unwrap();
         let title = m.call_method1("determine_title", (description,)).unwrap();
         title.extract::<String>()
     })
@@ -583,7 +592,7 @@ pub fn determine_title(description: &str) -> String {
 
 pub fn iter_forge_instances() -> impl Iterator<Item = Forge> {
     let ret = Python::with_gil(|py| {
-        let m = py.import("breezy.forge").unwrap();
+        let m = py.import_bound("breezy.forge").unwrap();
         let f = m.getattr("iter_forge_instances").unwrap();
         let instances = f.call0().unwrap();
         instances
@@ -597,7 +606,7 @@ pub fn iter_forge_instances() -> impl Iterator<Item = Forge> {
 
 pub fn create_project(name: &str, summary: Option<&str>) -> Result<(), Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.forge").unwrap();
+        let m = py.import_bound("breezy.forge").unwrap();
         m.call_method1("create_project", (name, summary))?;
         Ok(())
     })
@@ -605,7 +614,7 @@ pub fn create_project(name: &str, summary: Option<&str>) -> Result<(), Error> {
 
 pub fn get_proposal_by_url(url: &url::Url) -> Result<MergeProposal, Error> {
     Python::with_gil(|py| {
-        let m = py.import("breezy.forge").unwrap();
+        let m = py.import_bound("breezy.forge").unwrap();
         let proposal = m.call_method1("get_proposal_by_url", (url.to_string(),))?;
         Ok(MergeProposal::from(proposal.to_object(py)))
     })
