@@ -3,7 +3,7 @@ use pyo3::prelude::*;
 
 pub fn load_plugins() -> bool {
     Python::with_gil(|py| {
-        let m = py.import("breezy.plugin").unwrap();
+        let m = py.import_bound("breezy.plugin").unwrap();
         match m.call_method0("load_plugins") {
             Ok(_) => true,
             Err(e)
