@@ -1,17 +1,23 @@
 use crate::controldir::ControlDir;
 use crate::delta::TreeDelta;
 use crate::graph::Graph;
+use crate::location::AsLocation;
 use crate::revisionid::RevisionId;
 use crate::tree::RevisionTree;
+use chrono::DateTime;
+use chrono::TimeZone;
 use pyo3::exceptions::PyStopIteration;
 use pyo3::prelude::*;
 use pyo3::types::PyDict;
-use chrono::DateTime;
-use chrono::TimeZone;
 
-#[derive(Clone)]
 pub struct RepositoryFormat(PyObject);
 
+impl Clone for RepositoryFormat {
+    fn clone(&self) -> Self {
+        Python::with_gil(|py| RepositoryFormat(self.0.clone_ref(py)))
+    }
+}
+
 impl RepositoryFormat {
     pub fn supports_chks(&self) -> bool {
         Python::with_gil(|py| {
@@ -24,9 +30,14 @@ impl RepositoryFormat {
     }
 }
 
-#[derive(Clone)]
 pub struct Repository(PyObject);
 
+impl Clone for Repository {
+    fn clone(&self) -> Self {
+        Python::with_gil(|py| Repository(self.0.clone_ref(py)))
+    }
+}
+
 #[derive(Debug)]
 pub struct Revision {
     pub revision_id: RevisionId,
@@ -39,14 +50,14 @@ pub struct Revision {
 
 impl Revision {
     pub fn datetime(&self) -> DateTime<chrono::FixedOffset> {
-        let tz = chrono::FixedOffset::east(self.timezone);
+        let tz = chrono::FixedOffset::east_opt(self.timezone).unwrap();
         tz.timestamp_opt(self.timestamp as i64, 0).unwrap()
     }
 }
 
 impl ToPyObject for Revision {
     fn to_object(&self, py: Python) -> PyObject {
-        let kwargs = PyDict::new(py);
+        let kwargs = PyDict::new_bound(py);
         kwargs.set_item("message", self.message.clone()).unwrap();
         kwargs
             .set_item("committer", self.committer.clone())
@@ -57,18 +68,18 @@ impl ToPyObject for Revision {
         kwargs
             .set_item("parent_ids", self.parent_ids.iter().collect::<Vec<_>>())
             .unwrap();
-        py.import("breezy.revision")
+        py.import_bound("breezy.revision")
             .unwrap()
             .getattr("Revision")
             .unwrap()
-            .call((), Some(kwargs))
+            .call((), Some(&kwargs))
             .unwrap()
             .to_object(py)
     }
 }
 
 impl FromPyObject<'_> for Revision {
-    fn extract(ob: &'_ PyAny) -> PyResult<Self> {
+    fn extract_bound(ob: &Bound<PyAny>) -> PyResult<Self> {
         Ok(Revision {
             revision_id: ob.getattr("revision_id")?.extract()?,
             parent_ids: ob.getattr("parent_ids")?.extract()?,
@@ -119,11 +130,23 @@ impl Repository {
         Repository(obj)
     }
 
+    pub fn get_user_url(&self) -> url::Url {
+        Python::with_gil(|py| {
+            self.0
+                .getattr(py, "user_url")
+                .unwrap()
+                .extract::<String>(py)
+                .unwrap()
+                .parse()
+                .unwrap()
+        })
+    }
+
     pub fn fetch(
         &self,
         other_repository: &Repository,
         stop_revision: Option<&RevisionId>,
-    ) -> PyResult<()> {
+    ) -> Result<(), crate::error::Error> {
         Python::with_gil(|py| {
             self.0.call_method1(
                 py,
@@ -137,7 +160,7 @@ impl Repository {
         })
     }
 
-    pub fn revision_tree(&self, revid: &RevisionId) -> PyResult<RevisionTree> {
+    pub fn revision_tree(&self, revid: &RevisionId) -> Result<RevisionTree, crate::error::Error> {
         Python::with_gil(|py| {
             let o = self.0.call_method1(py, "revision_tree", (revid.clone(),))?;
             Ok(RevisionTree(o))
@@ -186,11 +209,63 @@ impl Repository {
         })
     }
 
-    pub fn get_revision(&self, revision_id: &RevisionId) -> PyResult<Revision> {
+    pub fn get_revision(&self, revision_id: &RevisionId) -> Result<Revision, crate::error::Error> {
         Python::with_gil(|py| {
             self.0
                 .call_method1(py, "get_revision", (revision_id.clone(),))?
                 .extract(py)
         })
+        .map_err(|e| e.into())
+    }
+
+    // TODO: This should really be on ForeignRepository
+    pub fn lookup_bzr_revision_id(
+        &self,
+        revision_id: &RevisionId,
+    ) -> Result<(Vec<u8>,), crate::error::Error> {
+        Python::with_gil(|py| {
+            self.0
+                .call_method1(py, "lookup_bzr_revision_id", (revision_id.clone(),))?
+                .extract::<(Vec<u8>, PyObject)>(py)
+        })
+        .map_err(|e| e.into())
+        .map(|(v, _m)| (v,))
+    }
+}
+
+pub fn open(base: impl AsLocation) -> Result<Repository, crate::error::Error> {
+    Python::with_gil(|py| {
+        let o = py
+            .import_bound("breezy.repository")?
+            .getattr("Repository")?
+            .call_method1("open", (base.as_location(),))?;
+        Ok(Repository::new(o.into()))
+    })
+}
+
+#[cfg(test)]
+mod repository_tests {
+    use crate::controldir::ControlDirFormat;
+    #[test]
+    fn test_simple() {
+        let td = tempfile::tempdir().unwrap();
+        let _dir = crate::controldir::create_standalone_workingtree(
+            td.path(),
+            &ControlDirFormat::default(),
+        )
+        .unwrap();
+        let _repo = crate::repository::open(td.path()).unwrap();
+    }
+
+    #[test]
+    fn test_clone() {
+        let td = tempfile::tempdir().unwrap();
+        let _dir = crate::controldir::create_standalone_workingtree(
+            td.path(),
+            &ControlDirFormat::default(),
+        )
+        .unwrap();
+        let repo = crate::repository::open(td.path()).unwrap();
+        let _repo2 = repo.clone();
     }
 }
