@@ -1,3 +1,4 @@
+use crate::error::Error;
 use pyo3::prelude::*;
 use pyo3::types::PyDict;
 
@@ -36,10 +37,41 @@ impl Transport {
                 .unwrap()
         })
     }
+
+    pub fn has(&self, path: &str) -> Result<bool, Error> {
+        pyo3::Python::with_gil(|py| {
+            Ok(self
+                .0
+                .call_method1(py, "has", (path,))?
+                .extract::<bool>(py)
+                .unwrap())
+        })
+    }
+
+    pub fn ensure_base(&self) -> Result<(), Error> {
+        pyo3::Python::with_gil(|py| {
+            self.0.call_method0(py, "ensure_base")?;
+            Ok(())
+        })
+    }
+
+    pub fn create_prefix(&self) -> Result<(), Error> {
+        pyo3::Python::with_gil(|py| {
+            self.0.call_method0(py, "create_prefix")?;
+            Ok(())
+        })
+    }
+
+    pub fn clone(&self, path: &str) -> Result<Transport, Error> {
+        pyo3::Python::with_gil(|py| {
+            let o = self.0.call_method1(py, "clone", (path,))?;
+            Ok(Transport(o.to_object(py)))
+        })
+    }
 }
 
 impl FromPyObject<'_> for Transport {
-    fn extract(obj: &PyAny) -> PyResult<Self> {
+    fn extract_bound(obj: &Bound<PyAny>) -> PyResult<Self> {
         Ok(Transport(obj.to_object(obj.py())))
     }
 }
@@ -50,47 +82,18 @@ impl ToPyObject for Transport {
     }
 }
 
-#[derive(Debug)]
-pub enum Error {
-    Python(PyErr),
-}
-
-impl From<PyErr> for Error {
-    fn from(e: PyErr) -> Self {
-        Error::Python(e)
-    }
-}
-
-impl From<Error> for PyErr {
-    fn from(e: Error) -> Self {
-        match e {
-            Error::Python(e) => e,
-        }
-    }
-}
-
-impl std::fmt::Display for Error {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            Error::Python(e) => e.fmt(f),
-        }
-    }
-}
-
-impl std::error::Error for Error {}
-
 pub fn get_transport(
     url: &url::Url,
     possible_transports: Option<&mut Vec<Transport>>,
 ) -> Result<Transport, Error> {
     pyo3::Python::with_gil(|py| {
-        let urlutils = py.import("breezy.transport").unwrap();
-        let kwargs = PyDict::new(py);
+        let urlutils = py.import_bound("breezy.transport").unwrap();
+        let kwargs = PyDict::new_bound(py);
         kwargs.set_item(
             "possible_transports",
             possible_transports.map(|t| t.iter().map(|t| t.0.clone()).collect::<Vec<PyObject>>()),
         )?;
-        let o = urlutils.call_method("get_transport", (url.to_string(),), Some(kwargs))?;
+        let o = urlutils.call_method("get_transport", (url.to_string(),), Some(&kwargs))?;
         Ok(Transport(o.to_object(py)))
     })
 }
