@@ -1,17 +1,10 @@
 use crate::branch::{Branch, RegularBranch};
 use crate::controldir::ControlDir;
+use crate::error::Error;
 use crate::lock::Lock;
 use crate::revisionid::RevisionId;
-use pyo3::import_exception;
 use pyo3::prelude::*;
 
-import_exception!(breezy.commit, PointlessCommit);
-import_exception!(breezy.commit, NoWhoami);
-import_exception!(breezy.errors, NotBranchError);
-import_exception!(breezy.errors, DependencyNotPresent);
-import_exception!(breezy.errors, DivergedBranches);
-import_exception!(breezy.transport, NoSuchFile);
-
 pub type Path = std::path::Path;
 pub type PathBuf = std::path::PathBuf;
 
@@ -55,7 +48,7 @@ impl pyo3::ToPyObject for Kind {
 }
 
 impl pyo3::FromPyObject<'_> for Kind {
-    fn extract(ob: &pyo3::PyAny) -> pyo3::PyResult<Self> {
+    fn extract_bound(ob: &Bound<pyo3::PyAny>) -> pyo3::PyResult<Self> {
         let s: String = ob.extract()?;
         match s.as_str() {
             "file" => Ok(Kind::File),
@@ -91,8 +84,9 @@ pub enum TreeEntry {
 }
 
 impl FromPyObject<'_> for TreeEntry {
-    fn extract(ob: &PyAny) -> PyResult<Self> {
-        match ob.getattr("kind")?.extract()? {
+    fn extract_bound(ob: &Bound<PyAny>) -> PyResult<Self> {
+        let kind: String = ob.getattr("kind")?.extract()?;
+        match kind.as_str() {
             "file" => {
                 let executable: bool = ob.getattr("executable")?.extract()?;
                 let kind: Kind = ob.getattr("kind")?.extract()?;
@@ -130,79 +124,15 @@ impl FromPyObject<'_> for TreeEntry {
     }
 }
 
-#[derive(Debug)]
-pub enum Error {
-    NoSuchFile(PathBuf),
-    Other(PyErr),
-}
-
-impl std::error::Error for Error {}
-
-impl std::fmt::Display for Error {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            Error::NoSuchFile(path) => write!(f, "No such file: {}", path.to_string_lossy()),
-            Error::Other(e) => write!(f, "{}", e),
-        }
-    }
-}
-
-impl From<PyErr> for Error {
-    fn from(e: PyErr) -> Self {
-        Python::with_gil(|py| {
-            if e.is_instance_of::<NoSuchFile>(py) {
-                return Error::NoSuchFile(e.value(py).getattr("path").unwrap().extract().unwrap());
-            }
-            Error::Other(e)
-        })
-    }
-}
-
-impl From<Error> for PyErr {
-    fn from(e: Error) -> Self {
-        match e {
-            Error::NoSuchFile(path) => NoSuchFile::new_err(path.to_string_lossy().to_string()),
-            Error::Other(e) => e,
-        }
-    }
-}
-
-#[derive(Debug)]
-pub enum PullError {
-    DivergedBranches,
-    Other(PyErr),
-}
-
-impl From<PyErr> for PullError {
-    fn from(e: PyErr) -> Self {
-        Python::with_gil(|py| {
-            if e.is_instance_of::<DivergedBranches>(py) {
-                return PullError::DivergedBranches;
-            }
-            PullError::Other(e)
-        })
-    }
-}
-
-impl std::fmt::Display for PullError {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            PullError::DivergedBranches => write!(f, "Diverged branches"),
-            PullError::Other(e) => write!(f, "{}", e),
-        }
-    }
-}
-
-impl std::error::Error for PullError {}
-
 pub trait Tree: ToPyObject {
-    fn get_tag_dict(&self) -> Result<std::collections::HashMap<String, RevisionId>, PyErr> {
+    fn get_tag_dict(&self) -> Result<std::collections::HashMap<String, RevisionId>, Error> {
         Python::with_gil(|py| {
             let branch = self.to_object(py).getattr(py, "branch")?;
             let tags = branch.getattr(py, "tags")?;
             let tag_dict = tags.call_method0(py, "get_tag_dict")?;
             tag_dict.extract(py)
         })
+        .map_err(|e: PyErr| -> Error { e.into() })
     }
 
     fn get_file(&self, path: &Path) -> Result<Box<dyn std::io::Read>, Error> {
@@ -250,6 +180,15 @@ pub trait Tree: ToPyObject {
         })
     }
 
+    fn get_symlink_target(&self, path: &Path) -> Result<PathBuf, Error> {
+        Python::with_gil(|py| {
+            let target = self
+                .to_object(py)
+                .call_method1(py, "get_symlink_target", (path,))?;
+            target.extract(py).map_err(|e| e.into())
+        })
+    }
+
     fn get_parent_ids(&self) -> Result<Vec<RevisionId>, Error> {
         Python::with_gil(|py| {
             Ok(self
@@ -270,6 +209,16 @@ pub trait Tree: ToPyObject {
         })
     }
 
+    fn kind(&self, path: &Path) -> Result<Kind, Error> {
+        Python::with_gil(|py| {
+            self.to_object(py)
+                .call_method1(py, "kind", (path,))
+                .unwrap()
+                .extract(py)
+                .map_err(|e| e.into())
+        })
+    }
+
     fn is_versioned(&self, path: &Path) -> bool {
         Python::with_gil(|py| {
             self.to_object(py)
@@ -288,7 +237,7 @@ pub trait Tree: ToPyObject {
         require_versioned: Option<bool>,
     ) -> Result<Box<dyn Iterator<Item = Result<TreeChange, Error>>>, Error> {
         Python::with_gil(|py| {
-            let kwargs = pyo3::types::PyDict::new(py);
+            let kwargs = pyo3::types::PyDict::new_bound(py);
             if let Some(specific_files) = specific_files {
                 kwargs.set_item("specific_files", specific_files)?;
             }
@@ -324,13 +273,14 @@ pub trait Tree: ToPyObject {
                 }
             }
 
-            Ok(Box::new(TreeChangeIter(self.to_object(py).call_method(
-                py,
-                "iter_changes",
-                (other.to_object(py),),
-                Some(kwargs),
-            )?))
-                as Box<dyn Iterator<Item = Result<TreeChange, Error>>>)
+            Ok(
+                Box::new(TreeChangeIter(self.to_object(py).call_method_bound(
+                    py,
+                    "iter_changes",
+                    (other.to_object(py),),
+                    Some(&kwargs),
+                )?)) as Box<dyn Iterator<Item = Result<TreeChange, Error>>>,
+            )
         })
     }
 
@@ -360,7 +310,7 @@ pub trait Tree: ToPyObject {
     ) -> Result<Box<dyn Iterator<Item = Result<(PathBuf, bool, Kind, TreeEntry), Error>>>, Error>
     {
         Python::with_gil(|py| {
-            let kwargs = pyo3::types::PyDict::new(py);
+            let kwargs = pyo3::types::PyDict::new_bound(py);
             if let Some(include_root) = include_root {
                 kwargs.set_item("include_root", include_root)?;
             }
@@ -399,11 +349,11 @@ pub trait Tree: ToPyObject {
                 }
             }
 
-            Ok(Box::new(ListFilesIter(self.to_object(py).call_method(
+            Ok(Box::new(ListFilesIter(self.to_object(py).call_method_bound(
                 py,
                 "list_files",
                 (),
-                Some(kwargs),
+                Some(&kwargs),
             )?))
                 as Box<
                     dyn Iterator<Item = Result<(PathBuf, bool, Kind, TreeEntry), Error>>,
@@ -458,7 +408,7 @@ pub trait Tree: ToPyObject {
 pub trait MutableTree: Tree {
     fn lock_write(&self) -> Result<Lock, Error> {
         Python::with_gil(|py| {
-            let lock = self.to_object(py).call_method0(py, "lock_write").unwrap();
+            let lock = self.to_object(py).call_method0(py, "lock_write")?;
             Ok(Lock::from(lock))
         })
     }
@@ -487,6 +437,18 @@ pub trait MutableTree: Tree {
         })
         .map_err(|e| e.into())
     }
+
+    fn remove(&self, files: &[&std::path::Path]) -> Result<(), Error> {
+        Python::with_gil(|py| -> Result<(), PyErr> {
+            self.to_object(py).call_method1(
+                py,
+                "remove",
+                (files.iter().map(|p| p.to_path_buf()).collect::<Vec<_>>(),),
+            )?;
+            Ok(())
+        })
+        .map_err(|e| e.into())
+    }
 }
 
 pub struct RevisionTree(pub PyObject);
@@ -499,99 +461,66 @@ impl ToPyObject for RevisionTree {
 
 impl Tree for RevisionTree {}
 
-#[derive(Debug)]
-pub enum CommitError {
-    PointlessCommit,
-    NoWhoami,
-    Other(PyErr),
-}
-
-impl std::fmt::Display for CommitError {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            CommitError::PointlessCommit => write!(f, "Pointless commit"),
-            CommitError::NoWhoami => write!(f, "No whoami"),
-            CommitError::Other(e) => write!(f, "Other error: {}", e),
-        }
+impl Clone for RevisionTree {
+    fn clone(&self) -> Self {
+        Python::with_gil(|py| RevisionTree(self.0.clone_ref(py)))
     }
 }
 
-impl std::error::Error for CommitError {}
+impl RevisionTree {
+    pub fn repository(&self) -> crate::repository::Repository {
+        Python::with_gil(|py| {
+            let repository = self.to_object(py).getattr(py, "_repository").unwrap();
+            crate::repository::Repository::new(repository)
+        })
+    }
+
+    pub fn get_revision_id(&self) -> RevisionId {
+        Python::with_gil(|py| {
+            self.to_object(py)
+                .call_method0(py, "get_revision_id")
+                .unwrap()
+                .extract(py)
+                .unwrap()
+        })
+    }
+
+    pub fn get_parent_ids(&self) -> Vec<RevisionId> {
+        Python::with_gil(|py| {
+            self.to_object(py)
+                .call_method0(py, "get_parent_ids")
+                .unwrap()
+                .extract(py)
+                .unwrap()
+        })
+    }
+}
 
 pub struct WorkingTree(pub PyObject);
 
+impl Clone for WorkingTree {
+    fn clone(&self) -> Self {
+        Python::with_gil(|py| WorkingTree(self.0.clone_ref(py)))
+    }
+}
+
 impl ToPyObject for WorkingTree {
     fn to_object(&self, py: Python) -> PyObject {
         self.0.to_object(py)
     }
 }
 
-#[derive(Debug)]
-pub enum WorkingTreeOpenError {
-    NotBranchError(String),
-    DependencyNotPresent(String, String),
-    Other(PyErr),
-}
-
-impl From<PyErr> for WorkingTreeOpenError {
-    fn from(err: PyErr) -> Self {
+impl WorkingTree {
+    pub fn is_control_filename(&self, path: &Path) -> bool {
         Python::with_gil(|py| {
-            if err.is_instance_of::<NotBranchError>(py) {
-                let l = err
-                    .value(py)
-                    .getattr("path")
-                    .unwrap()
-                    .extract::<String>()
-                    .unwrap();
-                WorkingTreeOpenError::NotBranchError(l)
-            } else if err.is_instance_of::<DependencyNotPresent>(py) {
-                let l = err
-                    .value(py)
-                    .getattr("library")
-                    .unwrap()
-                    .extract::<String>()
-                    .unwrap();
-                let e = err
-                    .value(py)
-                    .getattr("error")
-                    .unwrap()
-                    .extract::<String>()
-                    .unwrap();
-                WorkingTreeOpenError::DependencyNotPresent(l, e)
-            } else {
-                WorkingTreeOpenError::Other(err)
-            }
+            self.to_object(py)
+                .call_method1(py, "is_control_filename", (path,))
+                .unwrap()
+                .extract(py)
+                .unwrap()
         })
     }
-}
 
-impl From<WorkingTreeOpenError> for PyErr {
-    fn from(err: WorkingTreeOpenError) -> Self {
-        match err {
-            WorkingTreeOpenError::NotBranchError(l) => NotBranchError::new_err((l,)),
-            WorkingTreeOpenError::DependencyNotPresent(d, e) => {
-                DependencyNotPresent::new_err((d, e))
-            }
-            WorkingTreeOpenError::Other(err) => err,
-        }
-    }
-}
-
-impl std::fmt::Display for WorkingTreeOpenError {
-    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
-        match self {
-            WorkingTreeOpenError::NotBranchError(l) => write!(f, "Not branch error: {}", l),
-            WorkingTreeOpenError::DependencyNotPresent(d, e) => {
-                write!(f, "Dependency not present: {} {}", d, e)
-            }
-            WorkingTreeOpenError::Other(e) => write!(f, "Other error: {}", e),
-        }
-    }
-}
-
-impl std::error::Error for WorkingTreeOpenError {}
-
-impl WorkingTree {
     /// Return the base path for this working tree.
     pub fn basedir(&self) -> PathBuf {
         Python::with_gil(|py| {
@@ -619,49 +548,51 @@ impl WorkingTree {
         })
     }
 
-    pub fn open(path: &Path) -> Result<WorkingTree, WorkingTreeOpenError> {
+    pub fn open(path: &Path) -> Result<WorkingTree, Error> {
         Python::with_gil(|py| {
-            let m = py.import("breezy.workingtree")?;
+            let m = py.import_bound("breezy.workingtree")?;
             let c = m.getattr("WorkingTree")?;
             let wt = c.call_method1("open", (path,))?;
             Ok(WorkingTree(wt.to_object(py)))
         })
     }
 
-    pub fn open_containing(path: &Path) -> Result<(WorkingTree, PathBuf), WorkingTreeOpenError> {
+    pub fn open_containing(path: &Path) -> Result<(WorkingTree, PathBuf), Error> {
         Python::with_gil(|py| {
-            let m = py.import("breezy.workingtree")?;
+            let m = py.import_bound("breezy.workingtree")?;
             let c = m.getattr("WorkingTree")?;
-            let (wt, p): (&PyAny, String) =
+            let (wt, p): (Bound<PyAny>, String) =
                 c.call_method1("open_containing", (path,))?.extract()?;
             Ok((WorkingTree(wt.to_object(py)), PathBuf::from(p)))
         })
     }
 
-    pub fn basis_tree(&self) -> Box<dyn Tree> {
+    pub fn basis_tree(&self) -> Result<crate::tree::RevisionTree, Error> {
         Python::with_gil(|py| {
-            let tree = self.to_object(py).call_method0(py, "basis_tree").unwrap();
-            Box::new(RevisionTree(tree))
+            let tree = self.to_object(py).call_method0(py, "basis_tree")?;
+            Ok(RevisionTree(tree))
         })
     }
 
-    pub fn revision_tree(&self, revision_id: &RevisionId) -> Box<dyn Tree> {
+    pub fn revision_tree(&self, revision_id: &RevisionId) -> Result<Box<RevisionTree>, Error> {
         Python::with_gil(|py| {
-            let tree = self
-                .to_object(py)
-                .call_method1(py, "revision_tree", (revision_id.to_object(py),))
-                .unwrap();
-            Box::new(RevisionTree(tree))
+            let tree = self.to_object(py).call_method1(
+                py,
+                "revision_tree",
+                (revision_id.to_object(py),),
+            )?;
+            Ok(Box::new(RevisionTree(tree)))
         })
     }
 
-    pub fn get_tag_dict(&self) -> Result<std::collections::HashMap<String, RevisionId>, PyErr> {
+    pub fn get_tag_dict(&self) -> Result<std::collections::HashMap<String, RevisionId>, Error> {
         Python::with_gil(|py| {
             let branch = self.to_object(py).getattr(py, "branch")?;
             let tags = branch.getattr(py, "tags")?;
             let tag_dict = tags.call_method0(py, "get_tag_dict")?;
             tag_dict.extract(py)
         })
+        .map_err(|e: PyErr| -> Error { e.into() })
     }
 
     pub fn abspath(&self, path: &Path) -> Result<PathBuf, Error> {
@@ -716,9 +647,9 @@ impl WorkingTree {
         allow_pointless: Option<bool>,
         committer: Option<&str>,
         specific_files: Option<&[&Path]>,
-    ) -> Result<RevisionId, CommitError> {
+    ) -> Result<RevisionId, Error> {
         Python::with_gil(|py| {
-            let kwargs = pyo3::types::PyDict::new(py);
+            let kwargs = pyo3::types::PyDict::new_bound(py);
             if let Some(committer) = committer {
                 kwargs.set_item("committer", committer).unwrap();
             }
@@ -730,7 +661,7 @@ impl WorkingTree {
             }
 
             let null_commit_reporter = py
-                .import("breezy.commit")
+                .import_bound("breezy.commit")
                 .unwrap()
                 .getattr("NullCommitReporter")
                 .unwrap()
@@ -740,22 +671,14 @@ impl WorkingTree {
 
             Ok(self
                 .to_object(py)
-                .call_method(py, "commit", (message,), Some(kwargs))
-                .map_err(|e| {
-                    if e.is_instance_of::<PointlessCommit>(py) {
-                        CommitError::PointlessCommit
-                    } else if e.is_instance_of::<NoWhoami>(py) {
-                        CommitError::NoWhoami
-                    } else {
-                        CommitError::Other(e)
-                    }
-                })?
+                .call_method_bound(py, "commit", (message,), Some(&kwargs))
+                .map_err(|e| -> Error { e.into() })?
                 .extract(py)
                 .unwrap())
         })
     }
 
-    pub fn last_revision(&self) -> Result<RevisionId, PyErr> {
+    pub fn last_revision(&self) -> Result<RevisionId, Error> {
         Python::with_gil(|py| {
             let last_revision = self.to_object(py).call_method0(py, "last_revision")?;
             Ok(RevisionId::from(last_revision.extract::<Vec<u8>>(py)?))
@@ -766,17 +689,17 @@ impl WorkingTree {
         &self,
         source: &dyn crate::branch::Branch,
         overwrite: Option<bool>,
-    ) -> Result<(), PullError> {
+    ) -> Result<(), Error> {
         Python::with_gil(|py| {
             let kwargs = {
-                let kwargs = pyo3::types::PyDict::new(py);
+                let kwargs = pyo3::types::PyDict::new_bound(py);
                 if let Some(overwrite) = overwrite {
                     kwargs.set_item("overwrite", overwrite).unwrap();
                 }
                 kwargs
             };
             self.to_object(py)
-                .call_method(py, "pull", (source.to_object(py),), Some(kwargs))
+                .call_method_bound(py, "pull", (source.to_object(py),), Some(&kwargs))
         })
         .map_err(|e| e.into())
         .map(|_| ())
@@ -806,7 +729,7 @@ pub struct TreeChange {
 
 impl ToPyObject for TreeChange {
     fn to_object(&self, py: Python) -> PyObject {
-        let dict = pyo3::types::PyDict::new(py);
+        let dict = pyo3::types::PyDict::new_bound(py);
         dict.set_item("path", &self.path).unwrap();
         dict.set_item("changed_content", self.changed_content)
             .unwrap();
@@ -820,8 +743,8 @@ impl ToPyObject for TreeChange {
 }
 
 impl FromPyObject<'_> for TreeChange {
-    fn extract(obj: &PyAny) -> PyResult<Self> {
-        fn from_bool(o: &PyAny) -> PyResult<bool> {
+    fn extract_bound(obj: &Bound<PyAny>) -> PyResult<Self> {
+        fn from_bool(o: &Bound<PyAny>) -> PyResult<bool> {
             if let Ok(b) = o.extract::<isize>() {
                 Ok(b != 0)
             } else {
@@ -829,21 +752,21 @@ impl FromPyObject<'_> for TreeChange {
             }
         }
 
-        fn from_opt_bool_tuple(o: &PyAny) -> PyResult<(Option<bool>, Option<bool>)> {
-            let tuple = o.extract::<(Option<&PyAny>, Option<&PyAny>)>()?;
+        fn from_opt_bool_tuple(o: &Bound<PyAny>) -> PyResult<(Option<bool>, Option<bool>)> {
+            let tuple = o.extract::<(Option<Bound<PyAny>>, Option<Bound<PyAny>>)>()?;
             Ok((
-                tuple.0.map(from_bool).transpose()?,
-                tuple.1.map(from_bool).transpose()?,
+                tuple.0.map(|o| from_bool(&o.as_borrowed())).transpose()?,
+                tuple.1.map(|o| from_bool(&o.as_borrowed())).transpose()?,
             ))
         }
 
         let path = obj.getattr("path")?;
-        let changed_content = from_bool(obj.getattr("changed_content")?)?;
+        let changed_content = from_bool(&obj.getattr("changed_content")?)?;
 
-        let versioned = from_opt_bool_tuple(obj.getattr("versioned")?)?;
+        let versioned = from_opt_bool_tuple(&obj.getattr("versioned")?)?;
         let name = obj.getattr("name")?;
         let kind = obj.getattr("kind")?;
-        let executable = from_opt_bool_tuple(obj.getattr("executable")?)?;
+        let executable = from_opt_bool_tuple(&obj.getattr("executable")?)?;
         let copied = obj.getattr("copied")?;
 
         Ok(TreeChange {
@@ -861,8 +784,8 @@ impl FromPyObject<'_> for TreeChange {
 pub struct MemoryTree(pub PyObject);
 
 impl ToPyObject for MemoryTree {
-    fn to_object(&self, _py: Python) -> PyObject {
-        self.0.clone()
+    fn to_object(&self, py: Python) -> PyObject {
+        self.0.clone_ref(py)
     }
 }
 
@@ -884,3 +807,23 @@ impl From<&dyn Branch> for MemoryTree {
 impl Tree for MemoryTree {}
 
 impl MutableTree for MemoryTree {}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use crate::controldir::{create_standalone_workingtree, ControlDirFormat};
+
+    #[test]
+    fn test_remove() {
+        let td = tempfile::tempdir().unwrap();
+        let wt = create_standalone_workingtree(td.path(), &ControlDirFormat::default()).unwrap();
+        let path = td.path().join("foo");
+        std::fs::write(&path, b"").unwrap();
+        wt.add(&[(std::path::Path::new("foo"))]).unwrap();
+        wt.commit("Initial commit", None, None, None).unwrap();
+        assert!(wt.has_filename(&path));
+        wt.remove(&[&path]).unwrap();
+        assert!(!wt.is_versioned(&path));
+        std::mem::drop(td);
+    }
+}
