@@ -1,9 +1,11 @@
+use pyo3::prelude::*;
+
 pub fn join_segment_parameters(
     url: &url::Url,
     parameters: std::collections::HashMap<String, String>,
 ) -> url::Url {
     pyo3::Python::with_gil(|py| {
-        let urlutils = py.import("breezy.urlutils").unwrap();
+        let urlutils = py.import_bound("breezy.urlutils").unwrap();
         urlutils
             .call_method1("join_segment_parameters", (url.to_string(), parameters))
             .unwrap()
@@ -30,7 +32,7 @@ pub fn split_segment_parameters(
     url: &url::Url,
 ) -> (url::Url, std::collections::HashMap<String, String>) {
     pyo3::Python::with_gil(|py| {
-        let urlutils = py.import("breezy.urlutils").unwrap();
+        let urlutils = py.import_bound("breezy.urlutils").unwrap();
         urlutils
             .call_method1("split_segment_parameters", (url.to_string(),))
             .unwrap()
